//
// Perspective view of a color cube using LookAt() and Frustum()
//
// Colors are assigned to each vertex and then the rasterizer interpolates
// those colors across the triangles.
//
// Mouse buttons select a rotation axis (X-, Y-, or Z-axis) and rotate the
// cube 2 degrees about that axis.
//
// The 'A' key causes the cube to rotate at 4 seconds per rotation.  The
// 'a' key stops the rotation.
//
// The 'X' key moves the camera one unit in the +x direction.  The 'x' key
// moves the camera one unit in the -x direction.  Similarly for the 'Y',
// 'y', 'Z', and 'z' keys.
//
// The space key resets everything to initial conditions.


#include "Angel.h"

#ifdef WIN32
#include <Windows.h>
#include <Winbase.h>
#else
#include <sys/time.h>
int GetTickCount()
{
   struct timeval tv;
   gettimeofday(&tv, NULL);
   return 1000 * tv.tv_sec + tv.tv_usec / 1000;
}
#endif

int timeStamp;
GLfloat revsPerSec = 0.25;


const int ESC = 033;   // Escape key.

typedef vec4  color4;
typedef vec4  point4;


//(6 faces)(2 triangles/face)(3 vertices/triangle)
const int NumVertices = 36;


int Index = 0;
point4 points[NumVertices];
color4 colors[NumVertices];

// Vertices of a unit cube centered at origin, sides aligned with axes
point4 vertices[8] = {
    point4( -1.0, -1.0, -1.0, 1.0 ),
    point4(  1.0, -1.0, -1.0, 1.0 ),
    point4(  1.0,  1.0, -1.0, 1.0 ),
    point4( -1.0,  1.0, -1.0, 1.0 ),
    point4( -1.0, -1.0,  1.0, 1.0 ),
    point4(  1.0, -1.0,  1.0, 1.0 ),
    point4(  1.0,  1.0,  1.0, 1.0 ),
    point4( -1.0,  1.0,  1.0, 1.0 )
};

// RGBA colors
color4 vertex_colors[8] = {
    color4( 0.0, 0.0, 0.0, 1.0 ),  // black
    color4( 1.0, 0.0, 0.0, 1.0 ),  // red
    color4( 1.0, 1.0, 0.0, 1.0 ),  // yellow
    color4( 0.0, 1.0, 0.0, 1.0 ),  // green
    color4( 0.0, 0.0, 1.0, 1.0 ),  // blue
    color4( 1.0, 0.0, 1.0, 1.0 ),  // magenta
    color4( 1.0, 1.0, 1.0, 1.0 ),  // white
    color4( 0.0, 1.0, 1.0, 1.0 )   // cyan
};


GLuint  model_view;  // model-view matrix uniform shader variable location
GLuint  projection;  // projection matrix uniform shader variable location


// Projection transformation parameters
const GLfloat LEFT = -2.0, RIGHT = 2.0;
const GLfloat BOTTOM = -2.0, TOP = 2.0;
const GLfloat Z_NEAR = 2.0, Z_FAR = 20.0;


// Enumerated type for coordinate axes.
enum coord
   {
      X = 0,
      Y = 1,
      Z = 2
   };

vec3 cameraPos(0.0, 0.0, 5.0);

// Cube angle and current axis of rotation.
vec3 theta(0.0, 0.0, 0.0);
coord axis = Z;


// Function prototypes

void idle(void);


//----------------------------------------------------------------------------
// quad generates two triangles for each face and assigns colors
//    to the vertices

void
quad( int a, int b, int c, int d )
{
    colors[Index] = vertex_colors[a]; points[Index] = vertices[a]; Index++;
    colors[Index] = vertex_colors[b]; points[Index] = vertices[b]; Index++;
    colors[Index] = vertex_colors[c]; points[Index] = vertices[c]; Index++;
    colors[Index] = vertex_colors[a]; points[Index] = vertices[a]; Index++;
    colors[Index] = vertex_colors[c]; points[Index] = vertices[c]; Index++;
    colors[Index] = vertex_colors[d]; points[Index] = vertices[d]; Index++;
}


//----------------------------------------------------------------------------
// generate 12 triangles: 36 vertices and 36 colors
void
colorcube()
{
    quad( 1, 0, 3, 2 );
    quad( 2, 3, 7, 6 );
    quad( 3, 0, 4, 7 );
    quad( 6, 5, 1, 2 );
    quad( 4, 5, 6, 7 );
    quad( 5, 4, 0, 1 );
}


//----------------------------------------------------------------------------
// OpenGL initialization
void
init()
{
    colorcube();

    // Create a vertex array object
    GLuint vao;
    glGenVertexArrays( 1, &vao );
    glBindVertexArray( vao );

    // Create and initialize a buffer object
    GLuint buffer;
    glGenBuffers( 1, &buffer );
    glBindBuffer( GL_ARRAY_BUFFER, buffer );
    glBufferData( GL_ARRAY_BUFFER, sizeof(points) + sizeof(colors),
        NULL, GL_STATIC_DRAW );
    glBufferSubData( GL_ARRAY_BUFFER, 0, sizeof(points), points );
    glBufferSubData( GL_ARRAY_BUFFER, sizeof(points), sizeof(colors), colors );

    // Load shaders and use the resulting shader program
    GLuint program = InitShader( "vshaderCubeview.glsl",
                                 "fshaderCubeview.glsl" );
    glUseProgram( program );

    // set up vertex arrays
    GLuint vPosition = glGetAttribLocation( program, "vPosition" );
    glEnableVertexAttribArray( vPosition );
    glVertexAttribPointer( vPosition, 4, GL_FLOAT, GL_FALSE, 0,
            BUFFER_OFFSET(0) );

    GLuint vColor = glGetAttribLocation( program, "vColor" ); 
    glEnableVertexAttribArray( vColor );
    glVertexAttribPointer( vColor, 4, GL_FLOAT, GL_FALSE, 0,
            BUFFER_OFFSET(sizeof(points)) );

    model_view = glGetUniformLocation( program, "model_view" );
    projection = glGetUniformLocation( program, "projection" );
    
    glEnable( GL_DEPTH_TEST );
    glClearColor( 0.25, 0.25, 0.25, 1.0 ); 
}


//----------------------------------------------------------------------------

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    point4  camera( cameraPos[X],
                 cameraPos[Y],
                 cameraPos[Z],
                 1.0 );
    point4  at( 0.0, 0.0, 0.0, 1.0 );
    vec4    up( 0.0, 1.0, 0.0, 0.0 );

    mat4  mv = LookAt( camera, at, up ) * RotateZ(theta[Z])
               * RotateY(theta[Y]) * RotateX(theta[X]);

    glUniformMatrix4fv( model_view, 1, GL_TRUE, mv );

    glDrawArrays( GL_TRIANGLES, 0, NumVertices );

    glutSwapBuffers();
}


//----------------------------------------------------------------------------

void
keyboard( unsigned char key, int x, int y )
{
    switch( key )
    {

    case ESC: 
    case 'q':
    case 'Q':
       exit( EXIT_SUCCESS );
       break;

    case 'g':
    case 'G':
       glutIdleFunc(idle);
       timeStamp = GetTickCount();
       break;

    case 's':
    case 'S':
       glutIdleFunc(NULL);
       break;

    case 'x': cameraPos[X] -= 1.0; break;
    case 'X': cameraPos[X] += 1.0; break;

    case 'y': cameraPos[Y] -= 1.0; break;
    case 'Y': cameraPos[Y] += 1.0; break;

    case 'z': cameraPos[Z] -= 1.0; break;
    case 'Z': cameraPos[Z] += 1.0; break;

    case ' ':  // reset values to their defaults
       glutIdleFunc(NULL);
       cameraPos = vec3(0.0, 0.0, 5.0);
       theta = vec3(0.0, 0.0, 0.0);
       axis = Z;
       break;
    }

    std::cout << cameraPos[X] << " "
              << cameraPos[Y] << " "
              << cameraPos[Z] << std::endl;

    glutPostRedisplay();
}


void mouse(int btn, int state, int x, int y)
{
   if(btn==GLUT_LEFT_BUTTON && state == GLUT_DOWN) axis = X;
   if(btn==GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) axis = Y;
   if(btn==GLUT_RIGHT_BUTTON && state == GLUT_DOWN) axis = Z;

   theta[axis] += 2.0;

   if( theta[axis] > 360.0 ) theta[axis] -= 360.0;

   glutPostRedisplay();
}


//----------------------------------------------------------------------------

void
reshape( int w, int h )
{
   mat4 p;
   GLfloat ratio;

   glViewport( 0, 0, w, h );

   if(w <= h)
   {
      ratio = (GLfloat) h / (GLfloat) w;
      p = Frustum( LEFT, RIGHT, ratio * BOTTOM, ratio * TOP,
                   Z_NEAR, Z_FAR);
   }
   else
   {
      ratio = (GLfloat) w / (GLfloat) h;
      p = Frustum( ratio * LEFT, ratio * RIGHT, BOTTOM, TOP,
                   Z_NEAR, Z_FAR);
   }

   /* Or, just use perspective(). */

   //p = Perspective(45.0, w/h, 2.0, 20.0);

   glUniformMatrix4fv( projection, 1, GL_TRUE, p );
}


//----------------------------------------------------------------------------

void
idle( void )
{
   int currentTime = GetTickCount();

   theta[axis] += 0.001 * (currentTime - timeStamp) * 360.0 * revsPerSec;
   timeStamp = currentTime;

   if ( theta[axis] > 360.0 ) {
      theta[axis] -= 360.0;
   }
    
    glutPostRedisplay();
}


//----------------------------------------------------------------------------

int
main( int argc, char **argv )
{
    glutInit( &argc, argv );
    glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
    glutInitWindowSize( 512, 512 );
    glutInitContextVersion( 3, 2 );
    glutInitContextProfile( GLUT_CORE_PROFILE );
    glutCreateWindow( "Color Cube Viewer" );

    glewExperimental = GL_TRUE;
    glewInit();

    init();

    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
    glutMouseFunc(mouse);
    glutReshapeFunc( reshape );

    glutMainLoop();

    return 0;
}
